@@ -842,3 +842,389 @@ plotControlDistributions <- function(sca_data, title="", type="density"){
       facet_wrap(~factor(term), scales="free", labeller = label_parsed)
   return(sc1)
 }
+
+
+#' Compare different kinds of standard errors
+#'
+#' @description
+#' se_compare() takes in a regression formula (with or without fixed effects),
+#' data, and the types of standard errors desired, including clustered,
+#' heteroskedasticity-consistent, and bootstrapped. It then returns a data
+#' frame with coefficient and standard error estimates for easy comparison and
+#' plotting.
+#'
+#' @param formula A string containing a regression formula, with or without
+#'                fixed effects.
+#' @param data A data frame containing the variables provided in `formula` and
+#'             any clustering variables passed to `cluster`.
+#' @param types A string or vector of strings specifying what types of
+#'              standard errors are desired. Defaults to "all".
+#'
+#'              The following types are supported for non-fixed effects models:
+#'
+#'                With clustering: "HC0, "HC1", "HC2", "HC3".
+#'
+#'                Without clustering: "iid" (i.e. normal standard errors),
+#'                                    "HC0, "HC1", "HC2", "HC3",
+#'                                    "HC4", "HC4m", "HC5",
+#'                                    "bootstrapped".
+#'
+#'              The following types are supported for fixed effects models:
+#'
+#'                With clustering: "CL_FE" (clustered by fixed effects, i.e.
+#'                                 the default standard errors reported by
+#'                                 `feols()` if no clusters are supplied), if
+#'                                 clusters are supplied then the conventional
+#'                                 clustered standard errors from `feols()` are
+#'                                 estimated for each clustering variable. Two-
+#'                                 way clustered standard errors are not
+#'                                 supported at this time.
+#'
+#'                Without clustering: "HC0, "HC1", "HC2", "HC3",
+#'                                    "HC4", "HC4m", "HC5",
+#'                                    "bootstrapped".
+#' @param cluster A string or vector of strings specifying variables present in
+#'                `data` to be used for clustering standard errors.
+#' @param clusteredOnly A boolean indicating whether only standard errors with
+#'                      clustering should be estimated, defaults to `FALSE`.
+#' @param fixedEffectsOnly A boolean indicating whether only standard errors for
+#'                         fixed effects models should be estimated, defaults to
+#'                         `FALSE`.
+#' @param bootSamples An integer or vector of integers indicating how many times
+#'                    the model should be estimated with a random subset of the
+#'                    data. If a vector then every combination of `bootSamples`
+#'                    and `bootSampleSize` are estimated.
+#' @param bootSampleSize An integer or vector of integers indicating how many
+#'                       observations are in each random subset of the data.
+#'                       If a vector then every combination of `bootSamples`
+#'                       and `bootSampleSize` are estimated.
+#'
+#' @return A data frame where row represents an independent variable in the
+#'         model and each column a type of standard error. Coefficient estimates
+#'         for each variable are also included (column `"estimate"` for
+#'         non-fixed effects model and column `"estimate_FE"` for fixed effects
+#'         models). Columns are automatically named to specify the standard
+#'         error type.
+#'
+#'         Some examples:
+#'
+#'          "iid" = normal standard errors, i.e. assuming homoskedasticity
+#'
+#'          "CL_FE" = standard errors clustered by fixed effects
+#'
+#'          "bootstrap_k8n300_FE" =  bootstrapped standard errors for a fixed
+#'                                   effects model where `bootSamples = 8` and
+#'                                   `bootSampleSize = 300`
+#'
+#'          "CL_Depth_ID_FE" = standard errors clustered by the variable
+#'                               "Depth_ID" for a model with fixed effects
+#'
+#'          "HC0_Sta_ID" = HC0 standard errors clustered by the variable
+#'                           "Sta_ID"
+#'
+#'          Note: for fixed effects models the "(Intercept)" row will be all
+#'          `NA` because the intercept is not reported by `feols()` when fixed
+#'          effects are present.
+#'
+#' @export
+#'
+#' @examples
+#'
+#' se_compare(formula = "Salnty ~ T_degC + ChlorA + O2Sat | Sta_ID",
+#'            data = bottles, types = "all", cluster = c("Depth_ID", "Sta_ID"),
+#'            fixedEffectsOnly = FALSE, bootSamples=c(4, 8, 10),
+#'            bootSampleSize=c(300, 500))
+#'
+#' se_compare(formula = "Salnty ~ T_degC + ChlorA + O2Sat", data = bottles,
+#'            types = "bootstrapped", bootSamples = c(8, 10),
+#'            bootSampleSize = c(300, 500))
+#'
+#' se_compare(formula = "Salnty ~ T_degC + ChlorA", data = bottles,
+#'            types = c("HC0", "HC1", "HC3"))
+#'
+se_compare <- function(formula, data, types="all", cluster=NULL,
+                       clusteredOnly=FALSE, fixedEffectsOnly=FALSE,
+                       bootSamples=NULL, bootSampleSize=NULL){
+
+  # Create objects that will store the standard errors
+  ses_CL <- NULL
+  ses_HC <- NULL
+  ses_other <- NULL
+
+  # Create the object we will eventually return
+  ses <- NULL
+
+  # If the formula contains a pipe then fixed effects are assumed to be
+  # present and models are estimated with feols() rather than lm()
+  if(grepl("|", formula, fixed=T)){
+
+    model_fe <- tryCatch(feols(as.formula(formula), data=data),
+                         error=function(cond){
+                           message("Fixed effects model estimation failed.",
+                                   cond)
+                           return(NULL)
+                         })
+
+      if(is.null(model_fe)){
+        message("Fixed effects model estimation failed.")
+      }
+      else{
+      # Add FE model coefficients to the matrix
+      ses <- cbind(ses, matrix(c("(Intercept)"=NA, model_fe$coefficients), ncol=1,
+                               dimnames=list(c("(Intercept)",
+                                               names(model_fe$coefficients)),
+                                             c("estimate_FE"))))
+
+      # Case when user wants to cluster by FEs (i.e. the default SEs reported
+      # by feols()) or bootstrap
+      if(!clusteredOnly){
+        types_other <- c("CL_FE","bootstrapped")
+
+        if(!"all" %in% types){
+
+          if(length(setdiff(types, types_other)!=0)){
+            warning(paste0(setdiff(types, types_other),
+                           " not a valid type for SEs in FE model, ignoring.",
+                           collapse="\n"))
+          }
+
+          types_other <- types[types %in% types_other]
+
+        }
+
+        # Get the default standard errors from feols() output
+        if("CL_FE" %in% types_other){
+          ses_other <- cbind(ses_other, "CL_FE"=c("(Intercept)"=NA,
+                                                  coeftable(model_fe)[,2]))
+        }
+
+        # Get bootstrapped SEs
+        if("bootstrapped" %in% types_other& !is.null(bootSamples) &
+           !is.null(bootSampleSize)){
+
+          n_x <- length(model_fe$coefficients)
+
+          if(length(bootSamples)==1 & length(bootSampleSize==1)){
+
+            samples <- bootSamples
+            sample_sizes <- bootSampleSize
+
+            boot <- se_boot(data=data, formula=formula, n_x=n_x,
+                            n_samples=bootSamples[[1]],
+                            sample_size=bootSampleSize[[1]])
+
+            if(!is.null(boot)){
+              ses_other <- cbind(ses_other, boot)
+
+              colnames(ses_other)[ncol(ses_other)] <- paste("bootstrap_", "k",
+                                                            samples, "n",
+                                                            sample_sizes,
+                                                            "_FE", sep="")
+            }
+          }
+          else{
+            samples <- rep(bootSamples, length(bootSampleSize))
+            sample_sizes <- sort(rep(bootSampleSize, length(bootSamples)))
+
+            boot <- mapply(FUN=se_boot, n_samples=samples,
+                           sample_size=sample_sizes,
+                           MoreArgs=list(data=data, formula=formula, n_x=n_x))
+
+            if(!is.null(boot)){
+
+              colnames(boot) <- paste("bootstrap_", "k", samples, "n",
+                                      sample_sizes, "_FE", sep="")
+
+              ses_other <- cbind(ses_other, boot)
+            }
+          }
+        }
+        # Attach bootstapped/default SEs to the object to be returned
+        ses <- cbind(ses, ses_other)
+      }
+
+      # Estimate clustered standard errors for FE model for variables other than
+      # the FEs
+      if(!is.null(cluster)){
+
+        if(length(setdiff(cluster, colnames(data))!=0)){
+          warning(paste0(setdiff(cluster, colnames(data)),
+                         " not a valid clustering variable, ignoring.",
+                         collapse="\n"))
+
+          cluster <- cluster[cluster %in% colnames(data)]
+        }
+
+        if(!"all" %in% types){
+          if(length(setdiff(types[!types %in% c("bootstrapped", "iid")],
+                            types_CL))!=0){
+            warning(paste0(setdiff(setdiff(types, types_CL),
+                                   c("bootstrapped", "iid")),
+                           " not a valid type for clustered SEs, ignoring.",
+                           collapse="\n"))
+          }
+        }
+
+        # Estimate standard errors clustered by each desired variable
+        ses_CL <- sapply(cluster, FUN=function(c){
+          (feols(as.formula(formula), data=data,
+                 cluster=data[c]))$coeftable[,2]})
+
+        # Label them nicely
+        labs <- c()
+        for(c in cluster){
+          labs <- c(labs, paste0("CL", "_", c, "_FE"))
+        }
+
+        colnames(ses_CL) <- labs
+
+        ses_CL <- rbind("(Intercept)"=NA, ses_CL)
+
+        ses <- cbind(ses, ses_CL)
+      }
+    }
+
+  }
+  # Case when a non-FE model is desired
+  if(!fixedEffectsOnly){
+
+    # Allocate objects to hold SEs
+    ses_other <- NULL
+    ses_HC <- NULL
+    ses_CL <- NULL
+
+    # If the formula has FEs remove them
+    if(grepl("|",formula, fixed=T)){
+      formula <- str_trim(str_split(formula, fixed("|"))[[1]][[1]])
+    }
+
+    # Estimate the non-FE model and get the coefficients
+    model <- lm(formula=as.formula(formula), data=data)
+
+    ses <- cbind(ses, matrix(model$coefficients, ncol=1,
+                             dimnames=list(c(names(model$coefficients)), c("estimate"))))
+
+    # Parse the user's desired SE types
+    if(!clusteredOnly){
+      types_HC <- c("HC0", "HC1", "HC2", "HC3", "HC4", "HC4m", "HC5")
+      types_other <- c("iid", "bootstrapped")
+
+      if(!"all" %in% types){
+
+        if(length(setdiff(types, c(types_HC, types_other))!=0)){
+          warning(paste0(setdiff(types, c(types_HC, types_other)),
+                         " not a valid type for SEs, ignoring.", collapse="\n"))
+        }
+
+        types_HC <- types[types %in% types_HC]
+        types_other <- types[types %in% types_other]
+
+      }
+      # Get the normal iid standard errors
+      if("iid" %in% types_other){
+        ses_other <- cbind(ses_other, "iid"=summary(model)$coefficients[,2])
+      }
+
+      # Get HC standard errors
+      ses_HC <- sapply(types_HC,
+                       function(x) coeftest(model, vcov.=vcovHC, type=x)[,2])
+
+      # Get bootstrapped standard errors
+      if("bootstrapped" %in% types_other & !is.null(bootSamples) &
+         !is.null(bootSampleSize)){
+        n_x <- length(model$coefficients)-1
+
+        if(length(bootSamples)==1 & length(bootSampleSize==1)){
+
+          samples <- bootSamples
+          sample_sizes <- bootSampleSize
+
+          boot <- se_boot(data=data, formula=formula, n_x=n_x,
+                          n_samples=bootSamples[[1]],
+                          sample_size=bootSampleSize[[1]])
+
+          if(!is.null(boot)){
+            ses_other <- cbind(ses_other, boot)
+
+            colnames(ses_other)[ncol(ses_other)] <- paste("bootstrap_", "k",
+                                                          samples, "n",
+                                                          sample_sizes, sep="")
+          }
+        }
+        else{
+          samples <- rep(bootSamples, length(bootSampleSize))
+          sample_sizes <- sort(rep(bootSampleSize, length(bootSamples)))
+
+          boot <- mapply(FUN=se_boot, n_samples=samples,
+                         sample_size=sample_sizes,
+                         MoreArgs=list(data=data, formula=formula, n_x=n_x))
+
+          if(!is.null(boot)){
+            colnames(boot) <- paste("bootstrap_", "k", samples, "n",
+                                    sample_sizes, sep="")
+
+            ses_other <- cbind(ses_other, boot)
+          }
+        }
+
+      }
+
+      # Attach SEs to the return object
+      ses <- cbind(ses, ses_other, ses_HC)
+    }
+
+    # Case when clustered SEs for non-FE model are desired
+    if(!is.null(cluster)){
+
+      if(length(setdiff(cluster, colnames(data))!=0)){
+        warning(paste0(setdiff(cluster, colnames(data)),
+                       " not a valid clustering variable, ignoring.",
+                       collapse="\n"))
+
+        cluster <- cluster[cluster %in% colnames(data)]
+      }
+
+      types_CL <- c("HC0", "HC1", "HC2", "HC3")
+
+      if(!"all" %in% types){
+
+        if(length(setdiff(types[!types %in% c("bootstrapped", "iid")],
+                          types_CL))!=0){
+          warning(paste0(setdiff(setdiff(types, types_CL),
+                                 c("bootstrapped", "iid")),
+                         " not a valid type for clustered SEs, ignoring.",
+                         collapse="\n"))
+        }
+
+        types_CL <- types[types %in% types_CL]
+      }
+
+      # Estimate and extract clustered SEs
+      if(length(types_CL)>0){
+
+        ses_CL <- sapply(cluster, FUN=function(c, types){
+          sapply(types, function(x){
+            coeftest(model, vcov.=vcovCL, type=x, cluster=data[c])[,2]
+          })
+        }, types=types_CL, simplify=F)
+
+        ses_CL <- do.call(cbind, ses_CL)
+
+        # Label with type and clustering variable
+        labs <- c()
+        for(c in cluster){
+          for(t in types_CL){
+            labs <- c(labs, paste0(t, "_", c))
+          }
+        }
+
+        colnames(ses_CL) <- labs
+      }
+
+      ses <- cbind(ses, ses_CL)
+    }
+  }
+
+  # Coerce matrix to a data frame and return
+  return(as.data.frame(apply(ses, FUN=unlist, MARGIN=2)))
+}
